import requests


def _extract_prices(raw_section):
    ret = {}

    rows = raw_section.split(';')
    for row in rows:
        if row == '':
            continue

        cols = row.split(',')
        if len(cols) in [0, 10]:
            continue

        symbol_id = cols[0]
        isin = cols[1]
        short_name = cols[2]
        full_name = cols[3]
        open_price = int(cols[5])
        close_price = int(cols[6])
        last_price = int(cols[7])
        count = int(cols[8])
        volume = int(cols[9])
        value = int(cols[10])
        low_price = int(cols[11])
        high_price = int(cols[12])
        yesterday = int(cols[13])
        eps = int(cols[14]) if cols[14] != '' else None
        base_volume = int(cols[15])
        visit_count = int(cols[16])
        flow = int(cols[17])
        group = int(cols[18])
        max_price = int(float(cols[19]))
        min_price = int(float(cols[20]))
        z = int(cols[21])
        yval = int(cols[22])

        ret[symbol_id] = {
            'symbol_id': symbol_id,
            'symbol_short_name': short_name,
            'symbol_long_name': full_name,
            'isin': isin,
            'open': open_price,
            'close': close_price,
            'last': last_price,
            'high': high_price,
            'low': low_price,
            'count': count,
            'volume': volume,
            'value': value,
            'yesterday': yesterday,
            'eps': eps,
            'base_volume': base_volume,
            'visits_count': visit_count,
            'flow': flow,
            'group_code': group,
            'max': max_price,
            'min': min_price,
            'z': z,
            'yval': yval,
            'buy_orders': [],
            'sell_orders': [],
        }

    return ret


def _extract_orders(raw_section, watch):
    ret = {}
    rows = raw_section.split(';')
    for row in rows:
        if row == '':
            continue

        cols = row.split(',')

        symbol_id = cols[0]
        rank = cols[1]
        sell_count = cols[2]
        buy_count = cols[3]
        buy_price = cols[4]
        sell_price = cols[5]
        buy_volume = cols[6]
        sell_volume = cols[7]

        ainfo = watch.get(symbol_id, None)
        if ainfo is None:
            continue
        ainfo['buy_orders'].append({
            'rank': int(rank),
            'count': int(buy_count),
            'price': int(buy_price),
            'volume': int(buy_volume),
        })
        ainfo['sell_orders'].append({
            'rank': int(rank),
            'count': int(sell_count),
            'price': int(sell_price),
            'volume': int(sell_volume),
        })

        ret[symbol_id] = ainfo

    return ret


def fetch_watch_price_data():
    r = requests.get('http://www.tsetmc.com/tsev2/data/MarketWatchInit.aspx?h=0&r=0', timeout=20, verify=False)
    r.raise_for_status()

    raw_data = r.text
    sections = raw_data.split('@')

    watch = _extract_prices(sections[2])
    watch = _extract_orders(sections[3], watch)
    refid = sections[4]

    return watch, refid


def fetch_watch_client_type_data():
    r = requests.get('http://www.tsetmc.com/tsev2/data/ClientTypeAll.aspx', timeout=20, verify=False)
    r.raise_for_status()
    raw_data = r.text

    ret = {}
    sections = raw_data.split(';')
    for section in sections:
        r = section.split(',')
        symbol_id = r[0]
        ret[symbol_id] = {
            'natural': {
                'buy_count': int(r[1]),
                'buy_volume': int(r[3]),
                'sell_count': int(r[5]),
                'sell_volume': int(r[7]),
            },
            'legal': {
                'buy_count': int(r[2]),
                'buy_volume': int(r[4]),
                'sell_count': int(r[6]),
                'sell_volume': int(r[8]),
            },
        }

    return ret


_STATS_TRADES_INDICES = {
    1: 'average_value_3_month',  # میانگین ارزش معاملات در 3 ماه گذشته
    2: 'average_value_12_month',  # میانگین ارزش معاملات در 12 ماه گذشته
    3: 'average_value_rank_3_month',  # رتبه ارزش معاملات در 3 ماه گذشته
    4: 'average_value_rank_12_month',  # رتبه ارزش معاملات در 12 ماه گذشته
    5: 'average_volume_3_month',  # میانگین حجم معاملات در 3 ماه گذشته
    6: 'average_volume_12_month',  # میانگین حجم معاملات در 12 ماه گذشته
    7: 'average_volume_rank_3_month',  # رتبه حجم معاملات در 3 ماه گذشته
    8: 'average_volume_rank_12_month',  # رتبه حجم معاملات در 12 ماه گذشته
    9: 'average_count_3_month',  # میانگین دفعات معاملات روزانه در 3 ماه گذشته
    10: 'average_count_12_month',  # میانگین دفعات معاملات روزانه در 12 ماه گذشته
    11: 'average_count_rank_3_month',  # رتبه دفعات معاملات روزانه در 3 ماه گذشته
    12: 'average_count_rank_12_month',  # رتبه دفعات معاملات روزانه در 12 ماه گذشته
    13: 'average_price_no_base_last_day',  # قیمت میانگین وزنی آخرین روز - بدون دخالت حجم مبنا
    14: 'average_price_with_base_last_day',  # قیمت میانگین وزنی آخرین روز - با دخالت حجم مبنا
    15: 'value_last_day',  # ارزش معاملات آخرین روز
    16: 'volume_last_day',  # حجم معاملات آخرین روز
    17: 'count_last_day',  # دفعات معاملات در آخرین روز
}

_STATS_NEGATIVE_DAYS_INDICES = {
    18: 'days_count_3_month',  # تعداد روزهای منفی در 3 ماه گذشته
    19: 'days_count_12_month',  # تعداد روزهای منفی در 12 ماه گذشته
    20: 'days_percent_3_month',  # درصد روزهای منفی در 3 ماه گذشته
    21: 'days_percent_12_month',  # درصد روزهای منفی در 12 ماه گذشته
    22: 'days_rank_3_month',  # رتبه روزهای منفی در 3 ماه گذشته
    23: 'days_rank_12_month',  # رتبه روزهای منفی در 12 ماه گذشته
}

_STATS_NO_TRADE_DAYS_INDICES = {
    24: 'days_count_3_month',  # روزهای بدون معامله در 3 ماه گذشته
    25: 'days_count_12_month',  # روزهای بدون معامله در 12 ماه گذشته
}

_STATS_POSITIVE_DAYS_INDICES = {
    26: 'days_count_3_month',  # تعداد روزهای مثبت در 3 ماه گذشته
    27: 'days_count_12_month',  # تعداد روزهای مثبت در 12 ماه گذشته
    28: 'days_percent_3_month',  # درصد روزهای مثبت در 3 ماه گذشته
    29: 'days_percent_12_month',  # درصد روزهای مثبت در 12 ماه گذشته
    30: 'days_rank_3_month',  # رتبه روزهای مثبت در 3 ماه گذشته
    31: 'days_rank_12_month',  # رتبه روزهای مثبت در 12 ماه گذشته
}

_STATS_WITH_TRADE_DAYS_INDICES = {
    32: 'days_count_3_month',  # روزهای با معامله در 3 ماه گذشته
    33: 'days_count_12_month',  # روزهای با معامله در 12 ماه گذشته
    34: 'days_rank_3_month',  # رتبه روزهای با معامله در 3 ماه گذشته
    35: 'days_rank_12_month',  # رتبه روزهای با معامله در 12 ماه گذشته
}

_STATS_COMPANY_VALUE_INDICES = {
    36: 'total_value',  # ارزش شرکت در آخرین روز
    37: 'total_value_rank',  # رتبه ارزش شرکت در آخرین روز
}

_STATS_OPEN_DAYS_INDICES = {
    38: 'days_count_3_month',  # تعداد روزهای باز در 3 ماه گذشته
    39: 'days_count_12_month',  # تعداد روزهای باز در 12 ماه گذشته
    40: 'days_percent_3_month',  # درصد روزهای باز در 3 ماه گذشته
    41: 'days_percent_12_month',  # درصد روزهای باز در 12 ماه گذشته
    42: 'days_rank_3_month',  # رتبه روزهای باز در 3 ماه گذشته
    43: 'days_rank_12_month',  # رتبه روزهای باز در 12 ماه گذشته
}

_STATS_CLOSED_DAYS_INDICES = {
    44: 'days_count_3_month',  # تعداد روزهای بسته در 3 ماه گذشته
    45: 'days_count_12_month',  # تعداد روزهای بسته در 12 ماه گذشته
    46: 'days_percent_3_month',  # درصد روزهای بسته در 3 ماه گذشته
    47: 'days_percent_12_month',  # درصد روزهای بسته در 12 ماه گذشته
    48: 'days_rank_3_month',  # رتبه روزهای بسته در 3 ماه گذشته
    49: 'days_rank_12_month',  # رتبه روزهای بسته در 12 ماه گذشته
}

_STATS_CLIENT_TYPE_INDICES = {
    50: 'natural_buy_average_volume_3_month',  # میانگین حجم خرید حقیقی در 3 ماه گذشته
    51: 'natural_buy_average_volume_12_month',  # میانگین حجم خرید حقیقی در 12 ماه گذشته
    52: 'natural_buy_average_volume_rank_3_month',  # رتبه حجم خرید حقیقی در 3 ماه گذشته
    53: 'natural_buy_average_volume_rank_12_month',  # رتبه حجم خرید حقیقی در 12 ماه گذشته
    54: 'legal_buy_average_volume_3_month',  # میانگین حجم خرید حقوقی در 3 ماه گذشته
    55: 'legal_buy_average_volume_12_month',  # میانگین حجم خرید حقوقی در 12 ماه گذشته
    56: 'legal_buy_average_volume_rank_3_month',  # رتبه حجم خرید حقوقی در 3 ماه گذشته
    57: 'legal_buy_average_volume_rank_12_month',  # رتبه حجم خرید حقوقی در 12 ماه گذشته
    58: 'natural_buy_average_count_3_month',  # میانگین تعداد خریدار حقیقی در 3 ماه گذشته
    59: 'natural_buy_average_count_12_month',  # میانگین تعداد خریدار حقیقی در 12 ماه گذشته
    60: 'natural_buy_average_count_rank_3_month',  # رتبه تعداد خریدار حقیقی در 3 ماه گذشته
    61: 'natural_buy_average_count_rank_12_month',  # رتبه تعداد خریدار حقیقی در 12 ماه گذشته
    62: 'legal_buy_average_count_3_month',  # میانگین تعداد خریدار حقوقی در 3 ماه گذشته
    63: 'legal_buy_average_count_12_month',  # میانگین تعداد خریدار حقوقی در 12 ماه گذشته
    64: 'legal_buy_average_count_rank_3_month',  # رتبه تعداد خریدار حقوقی در 3 ماه گذشته
    65: 'legal_buy_average_count_rank_12_month',  # رتبه تعداد خریدار حقوقی در 12 ماه گذشته
    66: 'total_buy_average_count_3_month',  # میانگین تعداد خریداران در 3 ماه گذشته
    67: 'total_buy_average_count_12_month',  # میانگین تعداد خریداران در 12 ماه گذشته
    68: 'total_buy_average_count_rank_3_month',  # رتبه تعداد خریداران در 3 ماه گذشته
    69: 'total_buy_average_count_rank_12_month',  # رتبه تعداد خریداران در 12 ماه گذشته
    70: 'natural_sell_average_volume_3_month',  # میانگین حجم فروش حقیقی در 3 ماه گذشته
    71: 'natural_sell_average_volume_12_month',  # میانگین حجم فروش حقیقی در 12 ماه گذشته
    72: 'natural_sell_average_volume_rank_3_month',  # رتبه حجم فروش حقیقی در 3 ماه گذشته
    73: 'natural_sell_average_volume_rank_12_month',  # رتبه حجم فروش حقیقی در 12 ماه گذشته
    74: 'legal_sell_average_volume_3_month',  # میانگین حجم فروش حقوقی در 3 ماه گذشته
    75: 'legal_sell_average_volume_12_month',  # میانگین حجم فروش حقوقی در 12 ماه گذشته
    76: 'legal_sell_average_volume_rank_3_month',  # رتبه حجم فروش حقوقی در 3 ماه گذشته
    77: 'legal_sell_average_volume_rank_12_month',  # رتبه حجم فروش حقوقی در 12 ماه گذشته
    78: 'natural_sell_average_count_3_month',  # میانگین تعداد فروشنده حقیقی در 3 ماه گذشته
    79: 'natural_sell_average_count_12_month',  # میانگین تعداد فروشنده حقیقی در 12 ماه گذشته
    80: 'natural_sell_average_count_rank_3_month',  # رتبه تعداد فروشنده حقیقی در 3 ماه گذشته
    81: 'natural_sell_average_count_rank_12_month',  # رتبه تعداد فروشنده حقیقی در 12 ماه گذشته
    82: 'legal_sell_average_count_3_month',  # میانگین تعداد فروشنده حقوقی در 3 ماه گذشته
    83: 'legal_sell_average_count_12_month',  # میانگین تعداد فروشنده حقوقی در 12 ماه گذشته
    84: 'legal_sell_average_volume_rank_3_month',  # رتبه تعداد فروشنده حقوقی در 3 ماه گذشته
    85: 'legal_sell_average_volume_rank_12_month',  # رتبه تعداد فروشنده حقوقی در 12 ماه گذشته
    86: 'total_sell_average_count_3_month',  # میانگین تعداد فروشندگان در 3 ماه گذشته
    87: 'total_sell_average_count_12_month',  # میانگین تعداد فروشندگان در 12 ماه گذشته
    88: 'total_sell_average_count_rank_3_month',  # رتبه تعداد فروشندگان در 3 ماه گذشته
    89: 'total_sell_average_count_rank_12_month',  # رتبه تعداد فروشندگان در 12 ماه گذشته
}


def fetch_watch_stats_data():
    r = requests.get('http://www.tsetmc.com/tsev2/data/InstValue.aspx?t=a', timeout=20, verify=False)
    r.raise_for_status()
    raw_data = r.text

    symbol_id = None
    ret = {}
    sections = raw_data.split(';')
    for section in sections:
        r = section.split(',')
        x = 0
        if len(r) == 3:
            symbol_id = r[0]
            x = 1

        index = int(r[x + 0])
        val = int(r[x + 1]) if '.' not in r[x + 1] else float(r[x + 1])
        if symbol_id not in ret:
            ret[symbol_id] = {
                'trades': {},
                'negative_days': {},
                'no_trade_days': {},
                'positive_days': {},
                'with_trade_days': {},
                'company_value': {},
                'open_days': {},
                'closed_days': {},
                'client_type': {},
            }

        sub_name = None
        indices_obj = None

        if 1 <= index < 18:
            sub_name = 'trades'
            indices_obj = _STATS_TRADES_INDICES
        elif 18 <= index < 24:
            sub_name = 'negative_days'
            indices_obj = _STATS_NEGATIVE_DAYS_INDICES
        elif 24 <= index < 26:
            sub_name = 'no_trade_days'
            indices_obj = _STATS_NO_TRADE_DAYS_INDICES
        elif 26 <= index < 32:
            sub_name = 'positive_days'
            indices_obj = _STATS_POSITIVE_DAYS_INDICES
        elif 32 <= index < 36:
            sub_name = 'with_trade_days'
            indices_obj = _STATS_WITH_TRADE_DAYS_INDICES
        elif 36 <= index < 38:
            sub_name = 'company_value'
            indices_obj = _STATS_COMPANY_VALUE_INDICES
        elif 38 <= index < 44:
            sub_name = 'open_days'
            indices_obj = _STATS_OPEN_DAYS_INDICES
        elif 44 <= index < 50:
            sub_name = 'closed_days'
            indices_obj = _STATS_CLOSED_DAYS_INDICES
        elif 50 <= index < 90:
            sub_name = 'client_type'
            indices_obj = _STATS_CLIENT_TYPE_INDICES
        else:
            continue

        ret[symbol_id][sub_name][indices_obj[index]] = val

    return ret


def fetch_watch_historical_data():
    r = requests.get('http://members.tsetmc.com/tsev2/data/ClosingPriceAll.aspx', timeout=20, verify=False)
    r.raise_for_status()
    raw_data = r.text

    symbol_id = None
    ret = {}
    sections = raw_data.split(';')
    for section in sections:
        r = section.split(',')
        x = 0
        if len(r) == 11:
            symbol_id = r[0]

            x = 1

        index = int(r[x + 0])
        if symbol_id not in ret:
            ret[symbol_id] = {}
        ret[symbol_id][index] = {
            'close': int(r[x + 1]),
            'last': int(r[x + 2]),
            'count': int(r[x + 3]),
            'volume': int(r[x + 4]),
            'value': int(r[x + 5]),
            'low': int(r[x + 6]),
            'high': int(r[x + 7]),
            'yesterday': int(r[x + 8]),
            'open': int(r[x + 9]),
        }

    return ret



from typing import List

from jdatetime import date as jdate

from .core import shareholder as shareholder_core
from .core import symbol as symbol_core
from .core.cache import MemoryCache
from .day_details import SymbolDayDetails
from .shareholder import SymbolMajorShareholder


class Symbol:
    def __init__(self, symbol_id):
        self.id = symbol_id

    def get_price_data(self):
        """
        اطلاعات قیمت لحظه‌ای
        """
        # todo
        raise NotImplementedError('sorry, price date of symbol has not been implemented yet :(')

    def get_details(self) -> dict:
        """
        شناسه
        """
        if MemoryCache.exists('symbol_details', self.id):
            return MemoryCache.fetch('symbol_details', self.id)

        raw_details = symbol_core.get_symbol_details(self.id)
        ret = {
            'isin': raw_details.get('کد 12 رقمی نماد', None),
            'short_isin': raw_details.get('کد 5 رقمی نماد', None),
            'short_name': raw_details.get('نماد فارسی', None),
            'long_name': raw_details.get('نماد 30 رقمی فارسی', None),
            'english_name': raw_details.get('نام لاتین شرکت', None),

            'company_isin': raw_details.get('کد 12 رقمی شرکت', None),
            'company_short_isin': raw_details.get('کد 4 رقمی شرکت', None),
            'company_name': raw_details.get('نام شرکت', None),

            'market_code': raw_details.get('کد تابلو', None),
            'market_name': raw_details.get('بازار', None),

            'group_code': raw_details.get('کد گروه صنعت', None),
            'group_name': raw_details.get('گروه صنعت', None),

            'subgroup_code': raw_details.get('کد زیر گروه صنعت', None),
            'subgroup_name': raw_details.get('زیر گروه صنعت', None),
        }

        for key in ret:
            if isinstance(ret[key], str):
                ret[key] = ret[key].strip()

        MemoryCache.store('symbol_details', self.id, ret)

        return ret

    def get_daily_history(self) -> List[dict]:
        """
        سابقه
        """
        if MemoryCache.exists('symbol_daily_history', self.id):
            return MemoryCache.fetch('symbol_daily_history', self.id)

        raw_history = symbol_core.get_daily_history(self.id)
        ret = []

        for raw_day in raw_history:
            ret.append({
                'date': jdate.fromgregorian(date=raw_day['date']),
                'count': raw_day['count'],
                'volume': raw_day['volume'],
                'value': raw_day['value'],
                'yesterday': raw_day['yesterday_price'],
                'last': raw_day['last_price'],
                'first': raw_day['first_price'],
                'close': raw_day['close_price'],
                'low': raw_day['low_price'],
                'high': raw_day['high_price'],
            })

        MemoryCache.store('symbol_daily_history', self.id, ret)

        return ret

    def get_day_details(self, jyear: int, jmonth: int, jday: int) -> SymbolDayDetails:
        """
        اطلاعات یک روز خاص در سابقه
        """
        return SymbolDayDetails(self, jyear, jmonth, jday)

    def get_client_type_history(self) -> List[dict]:
        """
        تاریخچه‌ی حقیقی حقوقی
        """
        if MemoryCache.exists('symbol_client_type_history', self.id):
            return MemoryCache.fetch('symbol_client_type_history', self.id)

        raw_history = shareholder_core.get_client_type_history(self.id)
        ret = []
        for raw_proportion in raw_history:
            t = raw_proportion[0]
            ret.append({
                'date': jdate.fromgregorian(year=int(t[:4]), month=int(t[4:6]), day=int(t[6:])),
                'natural': {
                    'buy_count': int(raw_proportion[1]),
                    'buy_volume': int(raw_proportion[5]),
                    'buy_value': int(raw_proportion[9]),
                    'sell_count': int(raw_proportion[3]),
                    'sell_volume': int(raw_proportion[7]),
                    'sell_value': int(raw_proportion[11]),
                },
                'legal': {
                    'buy_count': int(raw_proportion[2]),
                    'buy_volume': int(raw_proportion[6]),
                    'buy_value': int(raw_proportion[10]),
                    'sell_count': int(raw_proportion[4]),
                    'sell_volume': int(raw_proportion[8]),
                    'sell_value': int(raw_proportion[12]),
                },
            })

        MemoryCache.store('symbol_client_type_history', self.id, ret)

        return ret

    def get_major_shareholders(self) -> List[SymbolMajorShareholder]:
        """
        لیست سهامداران عمده
        """
        if MemoryCache.exists('symbol_major_shareholders', self.id):
            return MemoryCache.fetch('symbol_major_shareholders', self.id)

        company_isin = self.get_details()['company_isin']
        raw_majors = shareholder_core.get_major_shareholders(company_isin)
        ret = []
        for raw_major in raw_majors:
            ret.append(
                SymbolMajorShareholder(self.id,
                                       company_isin=company_isin,
                                       holder_name=raw_major['name'],
                                       holder_id=raw_major['id'],
                                       percentage=raw_major['percentage'],
                                       shares_count=raw_major['shares_count'],
                                       change=raw_major['change'])
            )

        MemoryCache.store('symbol_major_shareholders', self.id, ret)

        return ret

    def __str__(self):
        return f'Symbol > {self.id}'



import json
import math
import re

import requests
from bs4 import BeautifulSoup

_intraday_pattern = re.compile("var (?P<name>.*)=(?P<content>\[[^;]*\]);")


def _load_script_vars_from_web(symbol_id, year, month, day):
    d = f'{year:04}{month:02}{day:02}'
    script_variables = {}

    # get data from url
    daily_content = requests.get(f'http://cdn.tsetmc.com/Loader.aspx?ParTree=15131P&i={symbol_id}&d={d}',
                                 timeout=20, verify=False).text

    # find and get script tags
    all_scripts = BeautifulSoup(daily_content, 'lxml').find_all('script')
    first_script = str(all_scripts[-3])
    second_script = str(all_scripts[-2])
    third_script = str(all_scripts[-1])

    # extract first script variables
    matches = _intraday_pattern.findall(first_script)
    for match in matches:
        script_variables[match[0]] = json.loads(match[1].replace('\'', '"'))

    # extract second script variables
    matches = _intraday_pattern.findall(second_script)
    for match in matches:
        script_variables[match[0]] = json.loads(match[1].replace('\'', '"'))

    # extract third script variables
    matches = _intraday_pattern.findall(third_script)
    for match in matches:
        script_variables[match[0]] = json.loads(match[1].replace('\'', '"'))

    return script_variables


def _extract_symbol_details_data(script_vars):
    symbol_details = {
        'full_name': script_vars['InstSimpleData'][0],
        'short_name': script_vars['InstSimpleData'][1],
        'market_short_name': script_vars['InstSimpleData'][2],
        'market_full_name': script_vars['InstSimpleData'][3],
        'isin': script_vars['InstSimpleData'][7],
        'shares_count': script_vars['InstSimpleData'][8],
        'base_volume': script_vars['InstSimpleData'][9],
    }

    return symbol_details


def _extract_price_data(script_vars):
    # State Data
    state_data = []
    for isd in script_vars['InstrumentStateData']:
        state_data.append(isd[1:])

    last_state_i = 0
    next_state_change_time = state_data[0][0] if len(state_data) > 1 else math.inf
    last_state = state_data[0][1] if len(state_data) > 1 else 'A '  # todo: convert to enum

    # Price Data
    price_data = []
    for cpd in script_vars['ClosingPriceData']:
        continuous_time = int(cpd[12])
        while continuous_time >= next_state_change_time:
            if len(state_data) > (last_state_i + 1):
                last_state_i += 1
                next_state_change_time = state_data[last_state_i][0]
                last_state = state_data[last_state_i][1]  # todo: convert to enum
            else:
                next_state_change_time = math.inf

        price_data.append({
            't': continuous_time,
            'lst': int(cpd[2]),
            'y': int(cpd[5]),
            'o': int(cpd[4]),
            'c': int(cpd[3]),
            'h': int(cpd[6]),
            'l': int(cpd[7]),
            'cnt': int(cpd[8]),
            'v': int(cpd[9]),
            's': last_state,
        })

    price_data.sort(key=lambda x: x['t'])
    return price_data


def _extract_orders_data(script_vars):
    orders = []
    for bld in script_vars['BestLimitData']:
        order = {
            't': bld[0],
            'rank': int(bld[1]),
            'bcnt': int(bld[2]),
            'bv': int(bld[3]),
            'bp': int(bld[4]),
            'scnt': int(bld[7]),
            'sv': int(bld[6]),
            'sp': int(bld[5]),
        }

        orders.append(order)

    orders.sort(key=lambda x: x['t'])
    return orders


def _extract_trade_data(script_vars):
    trades = [
        {
            'order': int(x[0]),
            't': int(f'{x[1][:2]}{x[1][3:5]}{x[1][6:]}'),
            'v': int(x[2]),
            'p': int(x[3]),
        }
        for x in script_vars['IntraTradeData']
    ]

    return trades


def _extract_share_holders_data(script_vars):
    share_holders = []
    for shd in script_vars['ShareHolderData']:
        share_holders.append({
            'id': shd[0],
            'shares_count': shd[2],
            'percentage': shd[3],
            'name': shd[5],
        })

    previous_share_holders = []
    for shd in script_vars['ShareHolderDataYesterday']:
        previous_share_holders.append({
            'id': shd[0],
            'shares_count': shd[2],
            'percentage': shd[3],
            'name': shd[5],
        })

    return previous_share_holders, share_holders


def load_intraday_data(symbol_id, year, month, day):
    script_vars = _load_script_vars_from_web(symbol_id, year, month, day)

    symbol_details = _extract_symbol_details_data(script_vars)
    price_data = _extract_price_data(script_vars)

    if not price_data:
        raise ValueError('there is no data for this symbol in the specified date')

    previous_shareholders, shareholders = _extract_share_holders_data(script_vars)
    trades = _extract_trade_data(script_vars)
    orders_data = _extract_orders_data(script_vars)

    return {
        'symbol_details': symbol_details,
        'price_data': price_data,
        'previous_shareholders': previous_shareholders,
        'shareholders': shareholders,
        'trades': trades,
        'orders_data': orders_data,
    }



from typing import List

from .core.group import get_all_groups


class SymbolGroup:
    @staticmethod
    def get_all_groups() -> List[dict]:
        """
        گرفتن لیست گروه‌های صنعت
        """

        return get_all_groups()



_GROUPS = [
    {
        'code': '01',
        'name': 'زراعت و خدمات وابسته',
    },
    {
        'code': '02',
        'name': 'جنگلداري و ماهيگيري',
    },
    {
        'code': '10',
        'name': 'استخراج زغال سنگ',
    },
    {
        'code': '11',
        'name': 'استخراج نفت گاز و خدمات جنبي جز اکتشاف',
    },
    {
        'code': '13',
        'name': 'استخراج کانه هاي فلزي',
    },
    {
        'code': '14',
        'name': 'استخراج ساير معادن',
    },
    {
        'code': '15',
        'name': 'حذف شده- فرآورده‌هاي غذايي و آشاميدني',
    },
    {
        'code': '17',
        'name': 'منسوجات',
    },
    {
        'code': '19',
        'name': 'دباغي، پرداخت چرم و ساخت انواع پاپوش',
    },
    {
        'code': '20',
        'name': 'محصولات چوبي',
    },
    {
        'code': '21',
        'name': 'محصولات كاغذي',
    },
    {
        'code': '22',
        'name': 'انتشار، چاپ و تکثير',
    },
    {
        'code': '23',
        'name': 'فراورده هاي نفتي، كك و سوخت هسته اي',
    },
    {
        'code': '24',
        'name': 'حذف شده-مواد و محصولات شيميايي',
    },
    {
        'code': '25',
        'name': 'لاستيك و پلاستيك',
    },
    {
        'code': '26',
        'name': 'توليد محصولات كامپيوتري الكترونيكي ونوري',
    },
    {
        'code': '27',
        'name': 'فلزات اساسي',
    },
    {
        'code': '28',
        'name': 'ساخت محصولات فلزي',
    },
    {
        'code': '29',
        'name': 'ماشين آلات و تجهيزات',
    },
    {
        'code': '31',
        'name': 'ماشين آلات و دستگاه‌هاي برقي',
    },
    {
        'code': '32',
        'name': 'ساخت دستگاه‌ها و وسايل ارتباطي',
    },
    {
        'code': '33',
        'name': 'ابزارپزشکي، اپتيکي و اندازه‌گيري',
    },
    {
        'code': '34',
        'name': 'خودرو و ساخت قطعات',
    },
    {
        'code': '35',
        'name': 'ساير تجهيزات حمل و نقل',
    },
    {
        'code': '36',
        'name': 'مبلمان و مصنوعات ديگر',
    },
    {
        'code': '38',
        'name': 'قند و شكر',
    },
    {
        'code': '39',
        'name': 'شرکتهاي چند رشته اي صنعتي',
    },
    {
        'code': '40',
        'name': 'عرضه برق، گاز، بخاروآب گرم',
    },
    {
        'code': '41',
        'name': 'جمع آوري، تصفيه و توزيع آب',
    },
    {
        'code': '42',
        'name': 'محصولات غذايي و آشاميدني به جز قند و شكر',
    },
    {
        'code': '43',
        'name': 'مواد و محصولات دارويي',
    },
    {
        'code': '44',
        'name': 'محصولات شيميايي',
    },
    {
        'code': '45',
        'name': 'پيمانكاري صنعتي',
    },
    {
        'code': '46',
        'name': 'تجارت عمده فروشي به جز وسايل نقليه موتور',
    },
    {
        'code': '47',
        'name': 'خرده فروشي،باستثناي وسايل نقليه موتوري',
    },
    {
        'code': '49',
        'name': 'كاشي و سراميك',
    },
    {
        'code': '50',
        'name': 'تجارت عمده وخرده فروشي وسائط نقليه موتور',
    },
    {
        'code': '51',
        'name': 'حمل و نقل هوايي',
    },
    {
        'code': '52',
        'name': 'انبارداري و حمايت از فعاليتهاي حمل و نقل',
    },
    {
        'code': '53',
        'name': 'سيمان، آهك و گچ',
    },
    {
        'code': '54',
        'name': 'ساير محصولات كاني غيرفلزي',
    },
    {
        'code': '55',
        'name': 'هتل و رستوران',
    },
    {
        'code': '56',
        'name': 'سرمايه گذاريها',
    },
    {
        'code': '57',
        'name': 'بانكها و موسسات اعتباري',
    },
    {
        'code': '58',
        'name': 'ساير واسطه گريهاي مالي',
    },
    {
        'code': '59',
        'name': 'اوراق حق تقدم استفاده از تسهيلات مسكن',
    },
    {
        'code': '60',
        'name': 'حمل ونقل، انبارداري و ارتباطات',
    },
    {
        'code': '61',
        'name': 'حمل و نقل آبي',
    },
    {
        'code': '63',
        'name': 'فعاليت هاي پشتيباني و كمكي حمل و نقل',
    },
    {
        'code': '64',
        'name': 'مخابرات',
    },
    {
        'code': '65',
        'name': 'واسطه‌گري‌هاي مالي و پولي',
    },
    {
        'code': '66',
        'name': 'بيمه وصندوق بازنشستگي به جزتامين اجتماعي',
    },
    {
        'code': '67',
        'name': 'فعاليتهاي كمكي به نهادهاي مالي واسط',
    },
    {
        'code': '68',
        'name': 'صندوق سرمايه گذاري قابل معامله',
    },
    {
        'code': '69',
        'name': 'اوراق تامين مالي',
    },
    {
        'code': '70',
        'name': 'انبوه سازي، املاك و مستغلات',
    },
    {
        'code': '71',
        'name': 'فعاليت مهندسي، تجزيه، تحليل و آزمايش فني',
    },
    {
        'code': '72',
        'name': 'رايانه و فعاليت‌هاي وابسته به آن',
    },
    {
        'code': '73',
        'name': 'اطلاعات و ارتباطات',
    },
    {
        'code': '74',
        'name': 'خدمات فني و مهندسي',
    },
    {
        'code': '76',
        'name': 'اوراق بهادار مبتني بر دارايي فكري',
    },
    {
        'code': '77',
        'name': 'فعالبت هاي اجاره و ليزينگ',
    },
    {
        'code': '82',
        'name': 'فعاليت پشتيباني اجرائي اداري وحمايت كسب',
    },
    {
        'code': '90',
        'name': 'فعاليت هاي هنري، سرگرمي و خلاقانه',
    },
    {
        'code': '93',
        'name': 'فعاليتهاي فرهنگي و ورزشي',
    },
    {
        'code': '98',
        'name': 'گروه اوراق غيرفعال',
    },
    {
        'code': 'X1',
        'name': 'شاخص',
    },
]


def get_all_groups():
    return _GROUPS



import os
import pickle
from collections import defaultdict
from os import path

_cache_dir = os.getenv('TSETMC_API_CACHE_DIRECTORY', path.expanduser('~/.tsetmc-api/cache'))


class PersistentCache:
    @staticmethod
    def store(major, minor, data):
        final_cache_dir = path.join(_cache_dir, major)
        final_path = path.join(final_cache_dir, f'{minor}.pickle')
        os.makedirs(final_cache_dir, exist_ok=True)

        with open(final_path, 'wb') as fp:
            pickle.dump(data, fp)

    @staticmethod
    def fetch(major, minor):
        final_path = path.join(_cache_dir, major, f'{minor}.pickle')

        try:
            with open(final_path, 'rb') as fp:
                return pickle.load(fp)
        except FileNotFoundError:
            return None

    @staticmethod
    def exists(major, minor):
        final_path = path.join(_cache_dir, major, f'{minor}.pickle')
        return path.exists(final_path)

    @staticmethod
    def remove(major, minor):
        if PersistentCache.exists(major, minor):
            final_path = path.join(_cache_dir, major, f'{minor}.pickle')
            os.unlink(final_path)


class MemoryCache:
    _cache = defaultdict(dict)

    @staticmethod
    def store(major, minor, data):
        MemoryCache._cache[major][minor] = data

    @staticmethod
    def fetch(major, minor):
        return MemoryCache._cache[major].get(minor, None)

    @staticmethod
    def exists(major, minor):
        return minor in MemoryCache._cache[major]

    @staticmethod
    def remove(major, minor):
        if MemoryCache.exists(major, minor):
            del MemoryCache._cache[major][minor]

    @staticmethod
    def clear(self):
        MemoryCache._cache.clear()



from typing import List

from dateutil import tz as timezoneutil
from jdatetime import date, time

from .core import day_details as day_details_core
from .core.cache import PersistentCache
from .shareholder import SymbolMajorShareholder

_timezone = timezoneutil.gettz('Asia/Tehran')


def _i_to_time(i):
    if isinstance(i, str):
        x = i
    else:
        x = str(i)

    if len(x) != 5:
        h = int(x[:2])
        m = int(x[2:4])
        s = int(x[4:])
    else:
        h = int(x[:1])
        m = int(x[1:3])
        s = int(x[3:])

    return time(hour=h, minute=m, second=s, tzinfo=_timezone)


def _create_snapshot(price_data, orders_data, distinct=True):
    max_t = price_data['t']
    snapshot = {
        'time': max_t,
        'last': price_data['lst'],
        'yesterday': price_data['y'],
        'open': price_data['o'],
        'high': price_data['h'],
        'close': price_data['c'],
        'low': price_data['l'],
        'count': price_data['cnt'],
        'volume': price_data['v'],
        'state': price_data['s'],
    } if distinct else price_data

    buy_orders = []
    sell_orders = []
    for to in orders_data:
        if to is not None:
            buy_orders.append({
                'time': _i_to_time(to['t']),
                'count': to['bcnt'],
                'volume': to['bv'],
                'price': to['bp'],
            })
            sell_orders.append({
                'time': _i_to_time(to['t']),
                'count': to['scnt'],
                'volume': to['sv'],
                'price': to['sp'],
            })

            max_t = max(max_t, to['t'])

    snapshot['time'] = _i_to_time(max_t)
    snapshot['buy_orders'] = buy_orders
    snapshot['sell_orders'] = sell_orders

    return snapshot


def _generate_snapshot_data(price_data, orders_data):
    max_pdi = len(price_data)
    max_odi = len(orders_data)
    pdi = 0
    odi = -1
    last_price_data = price_data[pdi]
    last_orders_data = [None, None, None]

    while True:
        phase = -1
        if pdi + 1 < max_pdi and odi + 1 < max_odi:
            if price_data[pdi + 1]['t'] < orders_data[odi + 1]['t']:
                phase = 1
            elif price_data[pdi + 1]['t'] > orders_data[odi + 1]['t']:
                phase = 2
            else:
                phase = 3
        else:
            if pdi + 1 == max_pdi and odi + 1 == max_odi:
                break
            elif pdi + 1 < max_pdi:
                phase = 1
            elif odi + 1 < max_odi:
                phase = 2

        # step price data and use the old order data
        if phase == 1 or phase == 3:
            pdi += 1
            last_price_data = price_data[pdi]

        # step orders
        if phase == 2 or phase == 3:
            t = orders_data[odi + 1]['t']
            orders_found = [None, None, None]
            while True:
                if odi + 1 >= max_odi:
                    break

                order = orders_data[odi + 1]
                rank = min(order['rank'] - 1, 2)
                if orders_found[rank] is not None or order['t'] > t:
                    break

                odi += 1
                orders_found[rank] = order
                last_orders_data[rank] = order

        yield _create_snapshot(last_price_data, last_orders_data)


class SymbolDayDetails:
    def __init__(self, symbol, jyear: int, jmonth: int, jday: int):
        self.symbol = symbol
        self.date = date(year=jyear, month=jmonth, day=jday)

        self._load()

    def _load(self):
        cache_minor = f'{self.symbol.id}-{self.date.year}-{self.date.month}-{self.date.day}'
        if PersistentCache.exists('asset_day_details', cache_minor):
            f = PersistentCache.fetch('asset_day_details', cache_minor)
            self._final_shareholders = f['final_shareholders']
            self._initial_shareholders = f['initial_shareholders']
            self._trades = f['trades']
            self._snapshots = f['snapshots']
            return

        mdate = self.date.togregorian()
        data = day_details_core.load_intraday_data(self.symbol.id, mdate.year, mdate.month, mdate.day)
        symbol_details = self.symbol.get_details()

        self._final_shareholders = []
        for shdata in data['shareholders']:
            self._final_shareholders.append(SymbolMajorShareholder(self.symbol.id,
                                                                   company_isin=symbol_details['company_isin'],
                                                                   holder_name=shdata['name'],
                                                                   holder_id=shdata['id'],
                                                                   percentage=shdata['percentage'],
                                                                   shares_count=shdata['shares_count']))

        self._initial_shareholders = []
        for shdata in data['yesterday_shareholders']:
            self._initial_shareholders.append(SymbolMajorShareholder(self.symbol.id,
                                                                     company_isin=symbol_details['company_isin'],
                                                                     holder_name=shdata['name'],
                                                                     holder_id=shdata['id'],
                                                                     percentage=shdata['percentage'],
                                                                     shares_count=shdata['shares_count']))

        self._trades = []
        for tdata in data['trades']:
            self._trades.append({
                'order': tdata['order'],
                'volume': tdata['v'],
                'price': tdata['p'],
                'time': _i_to_time(tdata['t']),
            })

        self._snapshots = list(_generate_snapshot_data(data['price_data'], data['orders_data']))

        PersistentCache.store('asset_day_details', cache_minor, {
            'final_shareholders': self._final_shareholders,
            'initial_shareholders': self._initial_shareholders,
            'trades': self._trades,
            'snapshots': self._snapshots,
        })

    def get_final_major_shareholders(self) -> List[SymbolMajorShareholder]:
        """
        سهامداران عمده در پایان روز
        """
        return self._final_shareholders

    def get_initial_major_shareholders(self) -> List[SymbolMajorShareholder]:
        """
        سهامداران عمده در ابتدای روز
        """
        return self._initial_shareholders

    def get_trades(self) -> List[dict]:
        """
        معاملات
        """
        return self._trades

    def get_snapshots(self) -> List[dict]:
        """
        تمام لحظات سهم در طور روز
        """
        return self._snapshots

    def get_snapshots_by_time(self, hour: int, minute: int, second: int) -> List[dict]:
        """
        وضعیت سهم در زمان خاصی از روز
        """
        snapshots = []
        t = time(hour=hour, minute=minute, second=second, tzinfo=_timezone)
        last_snapshot = None
        for snapshot in self._snapshots:
            if snapshot['time'] > t:
                if not snapshots:
                    return [last_snapshot]

                return snapshots

            if snapshot['time'] == t:
                snapshots.append(snapshot)

            last_snapshot = snapshot

        if not snapshots:
            return [last_snapshot]

        return snapshots

    def get_ticks(self, tick_length: int):
        """
        وضعیت سهم در لحظات مختلف روز در بازه‌هایی با طول tick_length
        """
        # todo
        raise NotImplementedError('sorry, ticks are not supported yet :(')

    def __str__(self):
        return f'AssetDayDetails > {self.symbol.id} @ {self.date}'



import locale
from datetime import date

import requests
from bs4 import BeautifulSoup


def get_client_type_history(symbol_id):
    raw = requests.get(f'http://www.tsetmc.com/tsev2/data/clienttype.aspx?i={symbol_id}',
                       timeout=20, verify=False).text
    raw_client_type_history = raw.split(';')

    ret = []
    for rsp in raw_client_type_history:
        ret.append(rsp.split(','))

    return ret


def get_major_shareholders(company_isin):
    raw = requests.get(f'http://www.tsetmc.com/Loader.aspx?Partree=15131T&c={company_isin}', timeout=20,
                       verify=False).text
    trs = BeautifulSoup(raw, 'lxml').find('tbody').find_all('tr')
    locale.setlocale(locale.LC_ALL, 'en_US.UTF8')

    ret = []
    for tr in trs:
        tds = tr.find_all('td')
        id = tr['onclick']
        id = id[id.index("'") + 1: id.index(",")]
        count = tds[1].find('div')['title']
        percentage = tds[2].contents[0]
        change = tds[3].contents[0]

        ret.append({
            'id': id,
            'name': str(tds[0].contents[0]),
            'shares_count': locale.atoi(count),
            'percentage': float(percentage),
            'change': locale.atoi(change),
        })

    return ret


def get_major_shareholder_details(company_isin, holder_id):
    raw = requests.get(f'http://www.tsetmc.com/tsev2/data/ShareHolder.aspx?i={holder_id}%2C{company_isin}',
                       timeout=20, verify=False).text
    raw = raw.split(';')

    ticks = []
    other_companies = []

    for z in raw:
        x = z.split(',')
        if len(x) == 2:
            ticks.append({
                'date': date(int(x[0][:4]), int(x[0][4:6]), int(x[0][6:])),
                'shares_count': int(x[1]),
            })
        elif len(x) == 4:
            other_companies.append({
                'asset_id': x[0][1:] if '#' in x[0] else x[0],
                'company_name': x[1],
                'shares_count': int(x[2]),
                'percentage': float(x[3]),
            })

    return {
        'activities': ticks,
        'other_assets': other_companies,
    }



from .group import SymbolGroup
from .symbol import Symbol
from .watch import Watch, Filter, AtiFilter, SahamFilter, WatchTick, SandoghFilter, KalaForoushFilter, \
    EkhtiarForoushFilter, HaghTaghaddomFilter, OraghMosharekatFilter, PayeFarabourseFilter



from typing import List

from jdatetime import date as jdate

from .core import shareholder as shareholder_core
from .core.cache import MemoryCache


class SymbolMajorShareholder:
    def __init__(self, symbol_id: str, company_isin: str, holder_name: str, holder_id: str, percentage: float,
                 shares_count: int, change: int = None):
        self.symbol_id = symbol_id
        self.company_isin = company_isin
        self.name = holder_name
        self.id = holder_id
        self.percentage = percentage
        self.shares_count = shares_count
        self.change = change

    def _get_details(self) -> dict:
        if MemoryCache.exists('major_shareholder_details', self.id):
            return MemoryCache.fetch('major_shareholder_details', self.id)
        raw = shareholder_core.get_major_shareholder_details(self.company_isin, self.id)

        ret = {
            'activities': [],
            'other_symbols': [],
        }
        for activity in raw['activities']:
            ret['activities'].append({
                'date': jdate.fromgregorian(date=activity['date']),
                'shares_count': activity['shares_count'],
            })

        from .symbol import Symbol
        for raw_symbol in raw['other_symbols']:
            ret['other_symbols'].append({
                'symbol': Symbol(raw_symbol['symbol_id']),
                'company_name': raw_symbol['company_name'],
                'shares_count': raw_symbol['shares_count'],
                'percentage': raw_symbol['percentage'],
            })

        MemoryCache.store('major_shareholder_details', self.id, ret)
        return ret

    def get_activity_history(self) -> List[dict]:
        """
        گرفتن سابقه‌ی فرد در این سهم (نمودار سهامدار عمده)
        """
        return self._get_details()['activities']

    def get_other_major_symbols(self) -> List[dict]:
        """
        سهم‌های دیگری که این فرد در آن‌ها هم سهامدار عمده است
        """
        return self._get_details()['other_symbols']

    def __str__(self):
        return f'SymbolMajorShareholder > {self.symbol_id}/{self.id} - {self.name}'



import time
from abc import ABC, abstractmethod
from traceback import print_exc

import schedule

from .core import watch as watch_core


class WatchTick:
    def __init__(self, price_data, client_type_data, historical_data, stats_data):
        self.price_data = price_data
        self.client_type_data = client_type_data
        self.historical_data = historical_data
        self.stats_data = stats_data

    def get_symbol_ids(self):
        """
        لیست آی دی همه‌ی نمادهای داخل دیده‌بان
        """
        return self.price_data.keys()

    def get_price_data(self, symbol_id):
        """
        اطلاعات قیمتی و سفارشات دیده‌بان
        """
        return self.price_data.get(symbol_id, {})

    def get_client_type_data(self, symbol_id):
        """
        اطلاعات حقیقی حقوقی دیده‌بان
        """
        return self.client_type_data.get(symbol_id, {})

    def get_historical_data(self, symbol_id):
        """
        اطلاعات سابقه‌ی دیده‌بان
        """
        return self.historical_data.get(symbol_id, {})

    def get_stats_data(self, symbol_id):
        """
        آمارهای کلیدی دیده‌بان
        """
        return self.stats_data.get(symbol_id, {})


class Filter(ABC):
    @abstractmethod
    def apply(self, symbol_id: str, watch_tick: WatchTick) -> bool:
        pass

    def __and__(self, other):
        me = self

        class AndFilter(Filter):
            def apply(self, symbol_id: str, watch_tick: WatchTick):
                return me.apply(symbol_id, watch_tick) and other.apply(symbol_id, watch_tick)

        return AndFilter()

    def __or__(self, other):
        me = self

        class OrFilter(Filter):
            def apply(self, symbol_id: str, watch_tick: WatchTick):
                return me.apply(symbol_id, watch_tick) or other.apply(symbol_id, watch_tick)

        return OrFilter()

    def __invert__(self):
        me = self

        class InvertedFilter(Filter):
            def apply(self, symbol_id: str, watch_tick: WatchTick):
                return not me.apply(symbol_id, watch_tick)

        return InvertedFilter()


class Watch:
    def __init__(self,
                 simple_refresh_time: int = 1,
                 client_type_refresh_time: int = 60,
                 ):
        self._price_data_rtime = simple_refresh_time
        self._client_type_data_rtime = client_type_refresh_time

        self._last_price_data = None
        self._last_client_type_data = None
        self._last_historical_data = None
        self._last_stats_data = None
        self._last_tick = None
        self._bindings = []

    def bind_listener(self, listener, filter_instance: Filter = None):
        self._bindings.append({
            'listener': listener,
            'filter': filter_instance,
        })

        return self

    def prepare_watch(self):
        """
        کنترل دستی
        """
        while self._last_historical_data is None:
            self._update_historical_data()

        while self._last_stats_data is None:
            self._update_stats_data()

        while self._last_price_data is None:
            self._update_price_data()

        while self._last_client_type_data is None:
            self._update_client_type_data()

        schedule.every(self._price_data_rtime).seconds.do(self._update_price_data)
        schedule.every(self._client_type_data_rtime).seconds.do(self._update_client_type_data)

    def step_watch(self):
        """
        کنترل دستی
        """
        schedule.run_pending()
        time.sleep(1)

    def start_watch(self):
        """
        شروع دیده‌بان
        """
        self.prepare_watch()
        while True:
            self.step_watch()

    def _publish(self):
        if self._last_historical_data is None:
            return

        if self._last_stats_data is None:
            return

        if self._last_price_data is None:
            return

        if self._last_client_type_data is None:
            return

        self._last_tick = WatchTick(price_data=self._last_price_data,
                                    client_type_data=self._last_client_type_data,
                                    historical_data=self._last_historical_data,
                                    stats_data=self._last_stats_data, )

        for binding in self._bindings:
            if binding['filter'] is None:
                binding['listener'](self._last_tick)
            else:
                binding['listener'](_create_filtered_tick(self._last_tick, binding['filter']))

    def _update_price_data(self):
        try:
            self._last_price_data, _ = watch_core.fetch_watch_price_data()
        except:
            print_exc()
            return

        self._publish()

    def _update_client_type_data(self):
        try:
            self._last_client_type_data = watch_core.fetch_watch_client_type_data()
        except Exception as ex:
            print_exc()
            return

        self._publish()

    def _update_historical_data(self):
        try:
            self._last_historical_data = watch_core.fetch_watch_historical_data()
        except:
            print_exc()

    def _update_stats_data(self):
        try:
            self._last_stats_data = watch_core.fetch_watch_stats_data()
        except:
            print_exc()


class SahamFilter(Filter):
    def apply(self, symbol_id: str, watch_tick: WatchTick) -> bool:
        yval = watch_tick.get_price_data(symbol_id=symbol_id).get('yval')
        return yval in [300, 303, 313]


class PayeFarabourseFilter(Filter):
    def apply(self, symbol_id: str, watch_tick: WatchTick) -> bool:
        yval = watch_tick.get_price_data(symbol_id=symbol_id).get('yval')
        return yval in [309]


class HaghTaghaddomFilter(Filter):
    def apply(self, symbol_id: str, watch_tick: WatchTick) -> bool:
        yval = watch_tick.get_price_data(symbol_id=symbol_id).get('yval')
        return yval in [400, 403, 404]


class OraghMosharekatFilter(Filter):
    def apply(self, symbol_id: str, watch_tick: WatchTick) -> bool:
        yval = watch_tick.get_price_data(symbol_id=symbol_id).get('yval')
        return yval in [306, 301, 706, 208]


class AtiFilter(Filter):
    def apply(self, symbol_id: str, watch_tick: WatchTick) -> bool:
        yval = watch_tick.get_price_data(symbol_id=symbol_id).get('yval')
        return yval in [263]


class SandoghFilter(Filter):
    def apply(self, symbol_id: str, watch_tick: WatchTick) -> bool:
        yval = watch_tick.get_price_data(symbol_id=symbol_id).get('yval')
        return yval in [305, 380]


class EkhtiarForoushFilter(Filter):
    def apply(self, symbol_id: str, watch_tick: WatchTick) -> bool:
        yval = watch_tick.get_price_data(symbol_id=symbol_id).get('yval')
        return yval in [600, 602, 605, 311, 312]


class KalaForoushFilter(Filter):
    def apply(self, symbol_id: str, watch_tick: WatchTick) -> bool:
        yval = watch_tick.get_price_data(symbol_id=symbol_id).get('yval')
        return yval in [308, 701]


def _create_filtered_tick(watch_tick: WatchTick, filter_instance: Filter):
    price_data = {}
    historical_data = {}
    stats_data = {}
    client_type_data = {}
    for symbol_id in watch_tick.get_symbol_ids():
        if filter_instance.apply(symbol_id, watch_tick):
            price_data[symbol_id] = watch_tick.get_price_data(symbol_id)
            client_type_data[symbol_id] = watch_tick.get_client_type_data(symbol_id)
            historical_data[symbol_id] = watch_tick.get_historical_data(symbol_id)
            stats_data[symbol_id] = watch_tick.get_stats_data(symbol_id)

    return WatchTick(price_data=price_data,
                     client_type_data=client_type_data,
                     historical_data=historical_data,
                     stats_data=stats_data)



from datetime import date

import requests
from bs4 import BeautifulSoup


def get_symbol_details(symbol_id):
    raw = requests.get(f'http://www.tsetmc.com/Loader.aspx?Partree=15131M&i={symbol_id}', timeout=20, verify=False).text

    ret = {}
    trs = BeautifulSoup(raw, 'lxml').find_all('tr')
    for tr in trs:
        tds = tr.find_all('td')
        ret[tds[0].contents[0]] = str(tds[1].contents[0])

    return ret


def get_daily_history(symbol_id):
    daily_content = requests.get(
        f'http://members.tsetmc.com/tsev2/data/InstTradeHistory.aspx?i={symbol_id}&Top=99999&A=0',
        timeout=20, verify=False).text
    raw_ticks = daily_content.split(';')

    ticks = []
    for raw_tick in raw_ticks:
        if raw_tick == '':
            continue

        tick_data = raw_tick.split('@')

        date_raw = tick_data[0]
        high_price = tick_data[1]
        low_price = tick_data[2]
        close_price = tick_data[3]
        last_price = tick_data[4]
        first_price = tick_data[5]
        yesterday_price = tick_data[6]
        value = tick_data[7]
        volume = tick_data[8]
        count = tick_data[9]

        ticks.append({
            'date': date(year=int(date_raw[:4]), month=int(date_raw[4:6]), day=int(date_raw[6:])),
            'first_price': int(first_price[:-3]),
            'high_price': int(high_price[:-3]),
            'low_price': int(low_price[:-3]),
            'close_price': int(close_price[:-3]),
            'last_price': int(last_price[:-3]),
            'yesterday_price': int(yesterday_price[:-3]),
            'value': int(float(value)),
            'volume': int(float(volume)),
            'count': int(float(count)),
        })

    return ticks
